# -*- coding: utf-8 -*-
"""
With the classes in this module the rather complicated calculation process
can be started with a simple method call.
Additionally, results are stored in a cache and can be reused later.
"""

import os
from .. import data
from .. import file
from ..file import File
from .algorithm import CavityCalculation, DomainCalculation, FakeDomainCalculation
from .discretization import DiscretizationCache, AtomDiscretization
from ...util import message
from ...config.configuration import config
from ...util.logger import Logger
from hashlib import sha256
import sys
from .. import bonds

__all__ = [
    "Calculation",
    "CalculationCache",
    "CalculationSettings",
    "calculated",
    "count_frames",
    "calculated_frames",
    "calculate_cavities",
    "getresults",
    "delete_center_cavity_information",
    "timestamp",
    "calculate",
]

logger = Logger("core.calculation")
logger.setstream("default", sys.stdout, Logger.WARNING)


class CalculationSettings(object):
    """
    Structure to store the parameters for one or more calculation.

    **Attributes:**
        `datasets` :
            Dictionary, which contains filenames (Strings) as keys.
            Each value is a list of Integers, which contains the frames.
            The value ``[-1]`` means 'all frames'.
        `resolution` :
            resolution parameter for the discretization
        `domains` :
            calculate cavitiy domains
        `surface_cavities` :
            calculate surface-based cavities
        `center_cavities` :
            calculate center-based cavities
        `recalculate` :
            results will be calculated even if cached results exists
        `exporthdf` :
            ``True`` if the results should be written into a hdf5 file.
            If ``False``, they are stored in the cache.
        `exporttext` :
            ``True`` if the results should be written into text files.
        `exportdir` :
            Directory for the exports. Only used, when at least one of
            the export parameters is ``True``. If `exportdir` is ``None``,
            the directory of the input file is used.
        `bonds` :
            calculate bonds
        `dihedral_angles` :
            calculate dihedral angles
    """

    def __init__(
        self,
        datasets,
        resolution=config.Computation.std_resolution,
        cutoff_radii=config.Computation.std_cutoff_radius,
        domains=False,
        surface_cavities=False,
        center_cavities=False,
        gyration_tensor=False,
        recalculate=False,
        exporthdf5=False,
        exporttext=False,
        exportdir=None,
    ):
        """ """
        self.datasets = datasets
        self.resolution = resolution
        self.cutoff_radii = cutoff_radii
        self.domains = domains
        self.surface_cavities = surface_cavities
        self.center_cavities = center_cavities
        self.gyration_tensor = gyration_tensor
        self.recalculate = recalculate
        self.exporthdf5 = exporthdf5
        self.exporttext = exporttext
        self.exportdir = exportdir
        self.bonds = False
        self.dihedral_angles = False

    def copy(self):
        """
        **Returns:**
            A deep copy of this object.
        """
        datasets = dict()
        for filename, frames in self.datasets.items():
            datasets[filename] = [f for f in frames]
        dup = self.__class__(datasets, self.resolution)
        dup.cutoff_radii = self.cutoff_radii
        dup.domains = self.domains
        dup.surface_cavities = self.surface_cavities
        dup.center_cavities = self.center_cavities
        dup.gyration_tensor = self.gyration_tensor
        dup.recalculate = self.recalculate
        dup.exporthdf5 = self.exporthdf5
        dup.exporttext = self.exporttext
        dup.exportdir = self.exportdir
        dup.bonds = self.bonds
        dup.dihedral_angles = self.dihedral_angles
        return dup


class Calculation(object):
    """
    This class provides the methods to start calculations.
    Optionally, results can be read from a cache.
    """

    def __init__(self, cachedir=None):
        """
        **Parameters:**
            `cachedir` :
                directory to store the cache files in; if none is given,
                a default one is used
        """
        if cachedir is None:
            cachedir = os.path.expanduser(config.Path.cache_dir)
        self.cachedir = cachedir
        max_cachefiles = config.Computation.max_cachefiles
        self.cache = CalculationCache(cachedir, max_cachefiles)

    def calculatedframes(self, filepath, resolution, surface=False, center=False):
        """
        Query the cache if it contains results for the given parameters.

        **Parameters:**
            `filepath` :
                absolute path of the input file
            `resolution` :
                resolution of the used discretization
            `surface` :
                query calculation results for surface-based cavities
            `center` :
                query calculation results for center-based cavities

        **Returns:**
            A :class:`core.data.TimestampList` containing the dates
            of the calculation or `None`.
        """
        info = None
        inputfile = File.open(filepath)
        # TODO: error handling
        if isinstance(inputfile, file.ResultFile):
            info = inputfile.info
        else:
            if filepath in self.cache:
                cf = self.cache[filepath]
                info = cf.info
        if info is not None:
            if surface:
                return info[resolution].surface_cavities
            elif center:
                return info[resolution].center_cavities
            else:
                return info[resolution].domains
        else:
            return data.TimestampList(inputfile.info.num_frames)

    def timestamp(self, filepath, frame, resolution, surface=False, center=False):
        """
        Query the cache if it contains results for the given parameters.

        **Returns:**
            The date and time when cached results were calculated
            on `None` if they do not exist.
        """
        calc = self.calculatedframes(filepath, resolution, surface, center)
        return calc[frame]

    def iscalculated(self, filepath, frame, resolution, surface=False, center=False):
        """
        **Returns:**
            If the cache contains results for the given parameters.
        """
        return not self.timestamp(filepath, frame, resolution, surface, center) is None

    def getresults(self, filepath, frame, resolution=None, surface=False, center=False):
        """
        Get cached results for the given parameters.

        **Parameters:**
            `filepath` :
                absolute path of the input file
            `frame` :
                the frame number
            `resolution` :
                resolution of the used discretization
            `surface` :
                query calculation results for surface-based cavities
            `center` :
                query calculation results for center-based cavities

        **Returns:**
            A :class:`core.data.Results` object if cached results exist,
            else `None`
        """
        inputfile = File.open(filepath)
        # TODO: error handling
        resultfile = None
        results = None
        if isinstance(inputfile, file.ResultFile):
            resultfile = inputfile
        elif filepath in self.cache:
            resultfile = self.cache[filepath]
        if resultfile is not None:
            if resolution is None:
                resolutions = sorted(resultfile.info.resolutions())[::-1]
                resolution = 64
                for res in resolutions:
                    if resultfile.info[res].domains[frame] is not None:
                        resolution = res
                        break
            results = resultfile.getresults(frame, resolution)
        if results is None:
            if resolution is None:
                resolution = 64
            results = data.Results(filepath, frame, resolution, inputfile.getatoms(frame), None, None, None)
        return results

    def calculateframe(
        self,
        filepath,
        frame,
        resolution,
        cutoff_radii=None,
        domains=False,
        surface=False,
        center=False,
        atoms=None,
        gyration_tensor_parameters=False,
        recalculate=False,
        last_frame=True,
    ):
        """
        Get results for the given parameters. They are either loaded from the
        cache or calculated.

        **Parameters:**
            `filepath` :
                absolute path of the input file
            `frame` :
                the frame number
            `resolution` :
                resolution of the used discretization
            `domains` :
                calculate cavitiy domains
            `cutoff_radii` :
                dict that maps element symbols to cutoff radii
            `surface` :
                calculate surface-based cavities
            `center` :
                calculate center-based cavities
            `gyration_tensor_parameters` :
                gyration tensor parameters will be calculated for cavities (they are always calculated for
                cavity domains)
            `recalculate` :
                results will be calculated even if cached results exists

        **Returns:**
            A :class:`core.data.Results` object.
        """
        # always recalculate if gyration tensor parameters shall be computed for center or surface based cavities
        recalculate = recalculate or (gyration_tensor_parameters and (center or surface))
        message.progress(0)
        inputfile = File.open(filepath)
        # TODO: error handling
        if isinstance(inputfile, file.ResultFile):
            resultfile = inputfile
        else:
            resultfile = self.cache[filepath]
        try:
            results = resultfile.getresults(frame, resolution)
        except Exception as e:
            logger.debug("error in resultfile.getresults: {}".format(e))
            results = None

        if atoms is None:
            atoms = inputfile.getatoms(frame)
        atoms.radii = cutoff_radii
        volume = atoms.volume
        if results is None:
            results = data.Results(filepath, frame, resolution, atoms, None, None, None)

        if recalculate:
            results.domains = None
            results.surface_cavities = None
            results.center_cavities = None

        if not (
            (domains and results.domains is None)
            or (surface and results.surface_cavities is None)
            or (center and results.center_cavities is None)
        ):
            message.print_message("Reusing results")
        else:
            cachepath = os.path.join(self.cachedir, "discretization_cache.hdf5")
            discretization_cache = DiscretizationCache(cachepath)
            with DiscretizationCache(cachepath) as discretization_cache:
                discretization = discretization_cache.get_discretization(volume, resolution)
            atom_discretization = AtomDiscretization(atoms, discretization)
            message.progress(10)  # TODO: improve the progress bar visualisation
            if (domains and results.domains is None) or (surface and results.surface_cavities is None):
                # CavityCalculation depends on DomainCalculation
                message.print_message("Calculating domains")
                domain_calculation = DomainCalculation(discretization, atom_discretization)
                if domain_calculation.critical_domains:
                    logger.warn(
                        "Found {:d} critical domains in file {}, frame {:d}. Domain indices: {}".format(
                            len(domain_calculation.critical_domains),
                            os.path.basename(filepath),
                            frame,
                            domain_calculation.critical_domains,
                        )
                    )
                    message.log(
                        "Found {:d} critical domains in file {}, frame {:d}".format(
                            len(domain_calculation.critical_domains),
                            os.path.basename(filepath),
                            frame + 1,
                        )
                    )
            if results.domains is None:
                results.domains = data.Domains(domain_calculation)
            message.progress(40)

            if surface and results.surface_cavities is None:
                message.print_message("Calculating surface-based cavities")
                cavity_calculation = CavityCalculation(
                    domain_calculation,
                    use_surface_points=True,
                    gyration_tensor_parameters=gyration_tensor_parameters,
                )
                results.surface_cavities = data.Cavities(cavity_calculation)
            message.progress(70)

            if center and results.center_cavities is None:
                message.print_message("Calculating center-based cavities")
                domain_calculation = FakeDomainCalculation(discretization, atom_discretization, results)  # equal, k, k
                cavity_calculation = CavityCalculation(
                    domain_calculation,
                    use_surface_points=False,
                    gyration_tensor_parameters=gyration_tensor_parameters,
                )
                results.center_cavities = data.Cavities(cavity_calculation)
            resultfile.addresults(results, overwrite=recalculate)

        message.progress(100)
        message.print_message("Calculation finished")
        if last_frame:
            message.finish()
        return results

    def calculate(self, calcsettings):
        """
        Calculate (or load from the cache) all results for the given settings.

        **Parameters:**
            `calcsettings` :
                :class:`CalculationSettings` object

        **Returns:**
            A list of list of :class:`core.data.Results` objects. The outer list contains
            an entry for each entry in `calcsettings.filenames`; the inner
            list has a `Results` entry for each frame specified in
            `calcsettings.frames`.
        """
        allresults = []
        for filename, frames in calcsettings.datasets.items():
            filepath = file.get_abspath(filename)
            fileprefix = os.path.basename(filename).rsplit(".", 1)[0]
            if calcsettings.exportdir is not None:
                exportdir = file.get_abspath(calcsettings.exportdir)
                # replace asterisks with directories
                dirlist = os.path.dirname(filepath).split("/")
                while "*" in exportdir:
                    i = exportdir.rindex("*")
                    exportdir = os.path.join(exportdir[:i], dirlist.pop() + exportdir[i + 1 :])
                if (calcsettings.exporthdf5 or calcsettings.exporttext) and not os.path.exists(exportdir):
                    os.makedirs(exportdir)
            else:
                exportdir = os.path.dirname(filepath)
            if calcsettings.exporthdf5:
                efpath = os.path.join(exportdir, fileprefix + ".hdf5")
                efpath = file.get_abspath(efpath)
                # copy atoms into HDF5 file
                exportfile = file.HDF5File.fromInputFile(efpath, filepath)
                # use HDF5 file as input
                filepath = efpath

            fileresults = []
            if frames[0] == -1:
                inputfile = File.open(filepath)
                frames = range(inputfile.info.num_frames)
            last_frame = False
            for frame in frames:
                # calculate single frame
                if frame is frames[-1]:
                    last_frame = True
                frameresult = self.calculateframe(
                    filepath,
                    frame,
                    calcsettings.resolution,
                    calcsettings.cutoff_radii,
                    domains=calcsettings.domains,
                    surface=calcsettings.surface_cavities,
                    center=calcsettings.center_cavities,
                    gyration_tensor_parameters=calcsettings.gyration_tensor,
                    recalculate=calcsettings.recalculate,
                    last_frame=last_frame,
                )
                # export to text file
                if calcsettings.exporttext:
                    fmt = os.path.join(exportdir, fileprefix) + "-{property}-{frame:06d}.txt"
                    if frameresult.atoms is not None:
                        frameresult.atoms.totxt(fmt.format(property="{property}", frame=frame + 1))
                    if frameresult.domains is not None:
                        try:
                            frameresult.domains.totxt(fmt.format(property="domain_{property}", frame=frame + 1))
                        except ValueError as e:
                            logger.warn(str(e))
                            logger.warn("The export of domain information could not be finished.")
                    if frameresult.surface_cavities is not None:
                        try:
                            frameresult.surface_cavities.totxt(
                                fmt.format(
                                    property="surface_cavities_{property}",
                                    frame=frame + 1,
                                )
                            )
                        except ValueError as e:
                            logger.warn(str(e))
                            logger.warn("The export of surface cavity information could not be finished.")
                    if frameresult.center_cavities is not None:
                        try:
                            frameresult.center_cavities.totxt(
                                fmt.format(
                                    property="center_cavities_{property}",
                                    frame=frame + 1,
                                )
                            )
                        except ValueError as e:
                            logger.warn(str(e))
                            logger.warn("The export of center cavity information could not be finished.")
                # gather results
                fileresults.append(frameresult)
            allresults.append(fileresults)
        return allresults


class CalculationCache(object):
    """
    Stores calculation results. Associates the input file with a
    'hdf5' file containing the calculated results.
    This is realized with a single directory containing hdf5 files that
    are named after the SHA256 value of the absolute path of the input file.
    Additionally, an index file "index.txt" is created, which contains
    the input file paths and the resulting hdf5 file names.
    """

    # TODO: replacement strategy
    def __init__(self, directory, max_cachefiles=0):
        """
        **Parameters:**
            `directory` :
                path of the directory in which the cahce files are stored
        """
        self.directory = directory
        self.max_cachefiles = max_cachefiles
        self.index = dict()
        self.indexfilepath = self.abspath("index.txt")
        if not os.path.isdir(directory):
            os.mkdir(directory)
        self.buildindex()
        self.writeindex()

    def __contains__(self, filepath):
        """
        **Parameters:**
            `filepath` :
                path to the input file

        **Returns:**
            If a cache file for the given input file exists.
        """
        sourcefilepath = file.get_abspath(filepath)
        cachefilepath = self.abspath(self.cachefile(sourcefilepath))
        return os.path.isfile(cachefilepath)

    def __getitem__(self, filepath):
        """
        Get the cache file for a given input file.

        **Parameters:**
            `filepath` :
                path to the input file

        **Returns:**
            A :class:`file.HDF5File` object.
            If no cache file exist for the input file, a new one is created.
        """
        sourcefilepath = file.get_abspath(filepath)
        cachefilepath = self.abspath(self.cachefile(sourcefilepath))
        if sourcefilepath not in self.index:
            self.index[sourcefilepath] = cachefilepath
            self.writeindex()
        cachefile = file.HDF5File(cachefilepath, sourcefilepath)
        # TODO: what if not sourcefilepath == cachefile.info.sourcefilepath
        return cachefile

    def abspath(self, filename):
        return os.path.abspath(os.path.join(file.get_abspath(self.directory), filename))

    def cachefile(self, filepath):
        return sha256(filepath.encode("utf-8")).hexdigest() + ".hdf5"

    def buildindex(self):
        self.index = dict()
        filenames = set(f for f in os.listdir(self.directory) if os.path.splitext(f)[1] == "hdf5")
        filenames.discard("discretization_cache.hdf5")
        for filename in filenames:
            cachepath = self.abspath(filename)
            try:
                cachefile = file.HDF5File(cachepath)
                filepath = cachefile.info.sourcefilepath
                if filepath is not None and filename == self.cachefile(filepath):
                    self.index[filepath] = filename
            except (IOError, AttributeError, RuntimeError):
                pass

    def cleanindex(self):
        if 0 < self.max_cachefiles < len(self.index):
            # get list of cachefiles and their last modification times (mtime)
            file_mtimes = []
            for filepath, cachefile in sorted(self.index.items()):
                cachepath = os.path.join(self.directory, cachefile)
                cache_mtime = os.path.getmtime(cachepath)
                file_mtimes.append((cache_mtime, cachepath))
            # delete older cachefiles so that only max_cachefiles are left
            file_mtimes.sort()
            for _, cachepath in file_mtimes[: -self.max_cachefiles]:
                os.remove(cachepath)
            # rebuild the index
            self.buildindex()

    def writeindex(self):
        self.cleanindex()
        with open(self.indexfilepath, "w") as f:
            for filepath, cachefile in sorted(self.index.items(), key=lambda x: x[0]):
                #  print >>f, filepath + "; " + cachefile
                print(f"{filepath}; {cachefile}", f)


calculation = Calculation()



# -*- coding: utf-8 -*-

from functools import partial
import itertools
from PySide6 import QtCore, QtWidgets
from ..object_select_widget import ObjectSelectWidget

counter = itertools.count(0, 1)


class ObjectTypeId:
    ATOM = next(counter)
    DOMAIN = next(counter)
    CENTER_BASED_CAVITY = next(counter)
    SURFACE_BASED_CAVITY = next(counter)


del counter


class ViewTabDock(QtWidgets.QDockWidget):
    """
    DockWidget for the 'view'-tab
    """

    def __init__(self, parent):
        QtWidgets.QDockWidget.__init__(self, "view", parent)
        self.setWidget(QtWidgets.QWidget())

        self.layout = QtWidgets.QHBoxLayout()
        self.view_tab = ViewTab(self.widget(), parent)

        self.layout.addWidget(self.view_tab)
        self.widget().setLayout(self.layout)

        self.setFeatures(QtWidgets.QDockWidget.DockWidgetMovable | QtWidgets.QDockWidget.DockWidgetFloatable)


class ViewTab(QtWidgets.QWidget):
    """
    tab 'view' in the main widget
    """

    def __init__(self, parent, main_window):
        QtWidgets.QWidget.__init__(self, parent)
        self.gl_widget = main_window.center.gl_widget
        self.vis_settings = self.gl_widget.vis.settings
        self.results = None
        self.init_gui()

    def init_gui(self):
        vbox = QtWidgets.QVBoxLayout()
        group_box = QtWidgets.QGroupBox("view settings", self)
        view_layout = QtWidgets.QVBoxLayout()

        # normal checkboxes are not used to prevent alignment problems between the different check box types
        self.box_check = ObjectSelectWidget("show bounding box", self, add_index_selector=False)
        self.atom_check = ObjectSelectWidget("show atoms", self)
        self.bonds_check = ObjectSelectWidget("show bonds", self, add_index_selector=False)
        self.domain_check = ObjectSelectWidget("show cavities (domains)", self)
        self.surface_cavity_check = ObjectSelectWidget("show cavities (surface method)", self)
        self.center_cavity_check = ObjectSelectWidget("show cavities (center method)", self)

        # synch with dataset
        self.box_check.setChecked(self.vis_settings.show_bounding_box)

        self.box_check.setChecked(True)
        self.atom_check.setChecked(True)
        self.bonds_check.setChecked(True)
        self.domain_check.setChecked(False)
        self.surface_cavity_check.setChecked(True)
        self.center_cavity_check.setChecked(False)

        self.box_check.state_changed.connect(partial(self.on_checkbox, self.box_check))
        self.atom_check.state_changed.connect(partial(self.on_checkbox, self.atom_check, clicked_box=ObjectTypeId.ATOM))
        self.bonds_check.state_changed.connect(partial(self.on_checkbox, self.bonds_check))
        self.domain_check.state_changed.connect(
            partial(self.on_checkbox, self.domain_check, clicked_box=ObjectTypeId.DOMAIN)
        )
        self.surface_cavity_check.state_changed.connect(
            partial(
                self.on_checkbox,
                self.surface_cavity_check,
                clicked_box=ObjectTypeId.SURFACE_BASED_CAVITY,
            )
        )
        self.center_cavity_check.state_changed.connect(
            partial(
                self.on_checkbox,
                self.center_cavity_check,
                clicked_box=ObjectTypeId.CENTER_BASED_CAVITY,
            )
        )

        self.atom_check.selection_indices_changed.connect(
            partial(self.object_indices_changed, object_type_id=ObjectTypeId.ATOM)
        )
        self.domain_check.selection_indices_changed.connect(
            partial(self.object_indices_changed, object_type_id=ObjectTypeId.DOMAIN)
        )
        self.surface_cavity_check.selection_indices_changed.connect(
            partial(
                self.object_indices_changed,
                object_type_id=ObjectTypeId.SURFACE_BASED_CAVITY,
            )
        )
        self.center_cavity_check.selection_indices_changed.connect(
            partial(
                self.object_indices_changed,
                object_type_id=ObjectTypeId.CENTER_BASED_CAVITY,
            )
        )

        self.domain_check.setDisabled(True)
        self.surface_cavity_check.setDisabled(True)
        self.center_cavity_check.setDisabled(True)
        self.update_cavity_buttons(self.results, None)

        view_layout.addWidget(self.box_check)
        view_layout.addWidget(self.atom_check)
        view_layout.addWidget(self.bonds_check)
        view_layout.addWidget(self.domain_check)
        view_layout.addWidget(self.surface_cavity_check)
        view_layout.addWidget(self.center_cavity_check)
        group_box.setLayout(view_layout)

        vbox.addWidget(group_box)
        vbox.addStretch()
        self.setLayout(vbox)
        self.show()

    def update_cavity_buttons(self, results, clicked_box=None):
        self.results = results

        if self.results is None:
            return

        if self.results.domains is None:
            self.domain_check.setDisabled(True)
            self.surface_cavity_check.setDisabled(True)
            self.center_cavity_check.setDisabled(True)
            return
        else:
            self.domain_check.setEnabled(True)

        # enable elements depending on, if they are computed
        self.surface_cavity_check.setEnabled(self.results.surface_cavities is not None)
        self.center_cavity_check.setEnabled(self.results.center_cavities is not None)

        # un- / check elements depending on current selection
        if clicked_box is None:
            return

        has_triggered_check_event = False
        if clicked_box == ObjectTypeId.SURFACE_BASED_CAVITY:
            is_checked = self.surface_cavity_check.isChecked()
            if is_checked:
                has_triggered_check_event = self.domain_check.isChecked() or self.center_cavity_check.isChecked()
                self.domain_check.setChecked(False)
                self.center_cavity_check.setChecked(False)
        elif clicked_box == ObjectTypeId.CENTER_BASED_CAVITY:
            is_checked = self.center_cavity_check.isChecked()
            if is_checked:
                has_triggered_check_event = self.surface_cavity_check.isChecked()
                self.surface_cavity_check.setChecked(False)
        elif clicked_box == ObjectTypeId.DOMAIN:
            is_checked = self.domain_check.isChecked()
            if is_checked:
                has_triggered_check_event = self.surface_cavity_check.isChecked()
                self.surface_cavity_check.setChecked(False)
        return has_triggered_check_event

    def on_checkbox(self, check_box, check_state, clicked_box=None):
        settings = self.gl_widget.vis.settings
        settings.show_bounding_box = self.box_check.isChecked()
        settings.show_atoms = self.atom_check.isChecked()
        if check_box == self.atom_check:
            if not settings.show_atoms:
                self.bonds_check.setEnabled(False)
                self.bonds_check.setChecked(False)
            else:
                self.bonds_check.setEnabled(True)
                if settings.show_bonds:
                    self.bonds_check.setChecked(True)
                else:
                    self.bonds_check.setChecked(False)
        if self.bonds_check.isEnabled():
            settings.show_bonds = self.bonds_check.isChecked()

        has_triggered_check_event = self.update_cavity_buttons(self.results, clicked_box)
        settings.show_domains = self.domain_check.isChecked()
        if check_box == self.domain_check and check_box.isChecked():
            settings.visible_domain_indices = check_box.indices
        settings.show_surface_cavities = self.surface_cavity_check.isChecked()
        if check_box == self.surface_cavity_check and check_box.isChecked():
            settings.visible_surface_cavity_indices = check_box.indices
        settings.show_center_cavities = self.center_cavity_check.isChecked()
        if check_box == self.center_cavity_check and check_box.isChecked():
            settings.visible_center_cavity_indices = check_box.indices
        # If no further check event was triggered, all settings are refreshed and the scene can be redrawn
        if not has_triggered_check_event:
            self.gl_widget.create_scene()
            self.window().updatestatus()

    def object_indices_changed(self, indices, object_type_id, update_scene=True):
        object_type_id2attribute_name = {
            ObjectTypeId.ATOM: "visible_atom_indices",
            ObjectTypeId.DOMAIN: "visible_domain_indices",
            ObjectTypeId.CENTER_BASED_CAVITY: "visible_center_cavity_indices",
            ObjectTypeId.SURFACE_BASED_CAVITY: "visible_surface_cavity_indices",
        }

        setattr(
            self.gl_widget.vis.settings,
            object_type_id2attribute_name[object_type_id],
            indices,
        )
        if update_scene:
            self.gl_widget.create_scene()



# -*- coding: utf-8 -*-
"""
To support various materials, different Bravais lattice systems need to be used
in pyMolDyn2, so that the shapes and periodic boundary conditions of different
crystalline structures are taken into account. These lattice systems are used
as simulation volume and this module provides a class for each of the 7 lattice
systems. All of these are centered at (0,0,0).

As monoclinic, orthorombic, tetragonal, rhombohedral and cubic systems are
special cases of a triclinic system, the corresponding classes inherit from
TriclinicVolume.

For more information about the Bravais lattice systems, see:
http://en.wikipedia.org/wiki/Bravais_lattice

Author: Florian Rhiem <f.rhiem@fz-juelich.de>
"""


from math import ceil, sin, cos, pi, sqrt, acos

cot = lambda alpha: cos(alpha) / sin(alpha)
import itertools
import numpy as np
import numpy.linalg as la

try:
    import numexpr as ne

    NUMEXPR = True
except ImportError:
    NUMEXPR = False


# TODO: check __str__ routines, make them use vectors


class HexagonalVolume(object):
    """
    A hexagonal volume centered in the origin with a side length of ``a`` (for
    the 6 individual outer sides) and a height of ``c``.
    """

    def __init__(self, a, c):
        self.a = float(a)
        self.c = float(c)
        f = 2 * self.a * sin(pi / 3)

        #: The lattice system translation vectors (`right`, `right-up`, `forward`)
        self.translation_vectors = [(cos(pi * i / 3) * f, sin(pi * i / 3) * f, 0) for i in range(2)] + [(0, 0, self.c)]

        #: The side lengths of an axis-aligned bounding box
        self.side_lengths = [2 * self.a * sin(pi / 3), 2 * self.a, self.c]

        #: The cell volume, calculated as 6 times the area of an equilateral
        #: triangle with side length a (3**0.5/4*a*a) times the height c.
        self.volume = 6 * (3**0.5 / 4 * self.a * self.a) * self.c

        #: A list of edges as point pairs
        self.edges = []
        for i in range(6):
            # edge on the lower hexagon
            p1 = (sin(pi / 3 * i) * self.a, cos(pi / 3 * i) * self.a, -0.5 * self.c)
            p2 = (
                sin(pi / 3 * (i + 1)) * self.a,
                cos(pi / 3 * (i + 1)) * self.a,
                -0.5 * self.c,
            )
            self.edges.append((p1, p2))
            # edge on the upper hexagon
            p3 = (sin(pi / 3 * i) * self.a, cos(pi / 3 * i) * self.a, 0.5 * self.c)
            p4 = (
                sin(pi / 3 * (i + 1)) * self.a,
                cos(pi / 3 * (i + 1)) * self.a,
                0.5 * self.c,
            )
            self.edges.append((p3, p4))
            # edge connecting the two hexagons
            p5 = (sin(pi / 3 * i) * self.a, cos(pi / 3 * i) * self.a, -0.5 * self.c)
            p6 = (sin(pi / 3 * i) * self.a, cos(pi / 3 * i) * self.a, 0.5 * self.c)
            self.edges.append((p5, p6))

    def is_inside(self, point):
        """
        True if ``point`` is inside of the volume, False otherwise.
        """
        if isinstance(point, np.ndarray) and len(point.shape) > 1:
            if NUMEXPR:
                a = self.a
                c = self.c
                sinpi3 = sin(pi / 3)
                cotpi3 = cot(pi / 3)
                x = point[:, 0]
                y = point[:, 1]
                z = point[:, 2]
                return ne.evaluate(
                    "(abs(z) <= c / 2) & \
                                    (abs(x) <= sinpi3 * a) & \
                                    (abs(y) <= a) & ((abs(y) <= a / 2) | \
                                    (abs(y) <= a - abs(x) * cotpi3))"
                )
            else:
                ap = np.abs(point)
                result = ap[:, 2] <= self.c / 2
                result = np.logical_and(result, ap[:, 0] <= sin(pi / 3) * self.a)
                result = np.logical_and(result, ap[:, 1] <= self.a)
                tmp = np.logical_or(ap[:, 1] <= self.a / 2, ap[:, 1] <= self.a - ap[:, 0] * cot(pi / 3))
                result = np.logical_and(result, tmp)
                return result
        else:
            x, y, z = point
            ax = abs(x)
            ay = abs(y)
            az = abs(z)
            if az > self.c / 2:
                return False
            if ax > sin(pi / 3) * self.a:
                return False
            if ay > self.a:
                return False
            if ay > self.a / 2 and ay > self.a - ax * cot(pi / 3):
                return False
            return True

    def get_equivalent_point(self, point):
        """
        For a given point, this method returns an equivalent point inside the volume.
        """
        equivalent_point = np.array(point)
        translation_vectors = np.array(self.translation_vectors)
        translation_vectors = np.append(translation_vectors, [translation_vectors[1] - translation_vectors[0]])
        translation_vectors.shape = (4, 3)
        translation_vector_lengths = np.array([la.norm(v) for v in translation_vectors])
        for i, translation_vector_length in enumerate(translation_vector_lengths):
            translation_vectors[i] /= translation_vector_length
        projection_matrix = np.matrix(translation_vectors)

        may_be_outside = True
        while may_be_outside:
            projected_point = projection_matrix * np.matrix(equivalent_point).T
            scaled_projected_point = np.array(projected_point.flat) / translation_vector_lengths
            max_index = np.argmax(np.abs(scaled_projected_point))
            if abs(scaled_projected_point[max_index]) > 0.5:
                may_be_outside = True
                equivalent_point -= (
                    ceil(scaled_projected_point[max_index] - 0.5)
                    * translation_vector_lengths[max_index]
                    * translation_vectors[max_index]
                )
            else:
                may_be_outside = False
        return tuple(equivalent_point)

    def _wrap(self, p):
        f = sqrt(3) * self.a
        M60 = np.matrix([[0.5, 0.5 * sqrt(3), 0.0], [-0.5 * sqrt(3), 0.5, 0.0], [0.0, 0.0, 1.0]])
        p[:, 2] -= np.ceil(p[:, 2] / self.c - 0.5) * self.c
        p[:, 0] -= np.ceil(p[:, 0] / f - 0.5) * f
        p = (M60 * p.T).T
        p[:, 0] -= np.ceil(p[:, 0] / f - 0.5) * f
        p = (M60 * p.T).T
        p[:, 0] -= np.ceil(p[:, 0] / f - 0.5) * f
        p = ((M60 * M60).T * p.T).T
        return p

    def get_distance(self, p1, p2):
        """
        Return the shortest distance vector between two points.

        **Parameters:**
            `p1`, `p2` :
                List or numpy array containing points as row vectors.

        **Returns:**
            Numpy array containing the distance vectors.
        """
        p1 = self._wrap(np.matrix(p1, copy=True))
        p2 = self._wrap(np.matrix(p2, copy=True))
        d = self._wrap(p2 - p1)
        return np.asarray(d)

    def __repr__(self):
        return "HEXAGONAL a=%f c=%f" % (self.a, self.c)

    def __str__(self):
        return "HEX %f %f" % (self.a, self.c)


class TriclinicVolume(object):
    """
    A triclinic volume centered at the origin with the angles ``alpha``,
    ``beta``, ``gamma`` and the side lengths ``a``, ``b`` and ``c``. It has the
    shape of a parallelepiped.
    """

    def __init__(self, *args):
        if len(args) == 6:
            a = args[0]
            b = args[1]
            c = args[2]
            alpha = args[3]
            beta = args[4]
            gamma = args[5]
            self.a = a
            self.b = b
            self.c = c
            self.alpha = alpha
            self.beta = beta
            self.gamma = gamma
            self.V = (
                1 - cos(alpha) ** 2 - cos(beta) ** 2 - cos(gamma) ** 2 + 2 * cos(alpha) * cos(beta) * cos(gamma)
            ) ** 0.5
            #: The cartesian-to-fractional transformation matrix
            self.M = np.matrix(
                [
                    [
                        1 / a,
                        1 / a * (-cos(gamma) / sin(gamma)),
                        1 / a * (cos(alpha) * cos(gamma) - cos(beta)) / (self.V * sin(gamma)),
                    ],
                    [
                        0,
                        1 / b * 1 / sin(gamma),
                        1 / b * (cos(beta) * cos(gamma) - cos(alpha)) / (self.V * sin(gamma)),
                    ],
                    [0, 0, 1 / c * sin(gamma) / self.V],
                ]
            )
            Minv = la.inv(self.M)
            #: The fracional-to-cartesian transformation matrix
            self.Minv = Minv
        else:
            v1 = np.array([float(f) for f in args[0:3]])
            v2 = np.array([float(f) for f in args[3:6]])
            v3 = np.array([float(f) for f in args[6:]])
            self.a = la.norm(v1)
            self.b = la.norm(v2)
            self.c = la.norm(v3)

            alpha = acos((v1.dot(v3)) / (self.a * self.c))
            beta = acos((v2.dot(v3)) / (self.b * self.c))
            gamma = acos((v1.dot(v2)) / (self.a * self.b))
            self.alpha = alpha
            self.beta = beta
            self.gamma = gamma

            self.V = (
                1 - cos(alpha) ** 2 - cos(beta) ** 2 - cos(gamma) ** 2 + 2 * cos(alpha) * cos(beta) * cos(gamma)
            ) ** 0.5
            self.Minv = np.matrix(np.array([v1, v2, v3]).T)
            self.M = np.matrix(la.inv(self.Minv))

        #: The lattice system translation vectors (`right`, `up`, `forward`)
        self.translation_vectors = [tuple(self.Minv.T[i].tolist()[0]) for i in range(3)]
        min_point = [float("inf")] * 3
        max_point = [float("-inf")] * 3
        for i, j, k in itertools.product((-0.5, 0, 0.5), repeat=3):
            point = self.Minv * np.matrix((i, j, k)).T
            point = point.T.tolist()[0]
            for l in range(3):
                min_point[l] = min(min_point[l], point[l])
                max_point[l] = max(max_point[l], point[l])

        #: The side lengths of an axis-aligned bounding box
        self.side_lengths = [d - c for c, d in zip(min_point, max_point)]

        #: The cell volume
        self.volume = self.V * self.a * self.b * self.c

        edges = [
            ((-0.5, -0.5, -0.5), (-0.5, -0.5, 0.5)),
            ((-0.5, -0.5, -0.5), (-0.5, 0.5, -0.5)),
            ((-0.5, -0.5, -0.5), (0.5, -0.5, -0.5)),
            ((-0.5, -0.5, 0.5), (-0.5, 0.5, 0.5)),
            ((-0.5, -0.5, 0.5), (0.5, -0.5, 0.5)),
            ((-0.5, 0.5, -0.5), (-0.5, 0.5, 0.5)),
            ((-0.5, 0.5, -0.5), (0.5, 0.5, -0.5)),
            ((-0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
            ((0.5, -0.5, -0.5), (0.5, -0.5, 0.5)),
            ((0.5, -0.5, -0.5), (0.5, 0.5, -0.5)),
            ((0.5, -0.5, 0.5), (0.5, 0.5, 0.5)),
            ((0.5, 0.5, -0.5), (0.5, 0.5, 0.5)),
        ]
        new_edges = []
        for edge in edges:
            point1, point2 = edge
            point1 = self.Minv * np.matrix(point1).T
            point1 = point1.T.tolist()[0]
            point2 = self.Minv * np.matrix(point2).T
            point2 = point2.T.tolist()[0]
            new_edges.append((point1, point2))
        #: A list of edges as point pairs
        self.edges = new_edges

    def is_inside(self, point):
        """
        Returns True if point is inside of the volume, False otherwise.
        """
        if isinstance(point, np.ndarray) and len(point.shape) > 1:
            fp = np.asarray((self.M * point.T).T)
            return np.all(np.abs(fp) < 0.5, axis=1)
        else:
            fractional_point = self.M * np.matrix(point).T
            return all((-0.5 < float(c) < 0.5 for c in fractional_point))

    def get_equivalent_point(self, point):
        """
        For a given point, this method returns an equivalent point inside the volume.
        """
        fractional_point = self.M * np.matrix(point).T
        fractional_point = fractional_point.T.tolist()[0]
        for i in range(3):
            fractional_point[i] -= ceil(fractional_point[i] - 0.5)
        new_point = self.Minv * np.matrix(fractional_point).T
        new_point = tuple(new_point.T.tolist()[0])
        return new_point

    def get_distance(self, p1, p2):
        """
        Return the shortest distance vector between two points.

        **Parameters:**
            `p1`, `p2` :
                List or numpy array containing points as row vectors.

        **Returns:**
            Numpy array containing the distance vectors.
        """
        tp1 = (self.M * np.matrix(p1, copy=False).T).T
        tp2 = (self.M * np.matrix(p2, copy=False).T).T
        td = tp2 - tp1
        td -= np.ceil(td - 0.5)
        return np.array((self.Minv * td.T).T, copy=False)

    def __repr__(self):
        return "TRICLINIC a=%f b=%f c=%f alpha=%f beta=%f gamma=%f" % (
            self.a,
            self.b,
            self.c,
            self.alpha,
            self.beta,
            self.gamma,
        )

    def __str__(self):
        return "TRI " + " ".join([str(x) for v in self.translation_vectors for x in v])


class MonoclinicVolume(TriclinicVolume):
    """
    A monoclinic volume, a special case of a triclinic volume with and ``alpha=gamma=pi/2``
    """

    def __init__(self, *args):
        if len(args) == 4:
            a = args[0]
            b = args[1]
            c = args[2]
            beta = args[3]
            alpha = pi / 2
            gamma = pi / 2
            TriclinicVolume.__init__(self, a, b, c, alpha, beta, gamma)
        else:
            v1 = np.array([float(f) for f in args[0:3]])
            v2 = np.array([float(f) for f in args[3:6]])
            v3 = np.array([float(f) for f in args[6:]])
            TriclinicVolume.__init__(self, v1, v2, v3)

    def __repr__(self):
        return "MONOCLINIC a=%f b=%f c=%f beta=%f" % (self.a, self.b, self.c, self.beta)

    def __str__(self):
        return "MON %f %f %f %f" % (self.a, self.b, self.c, self.beta)


class OrthorhombicVolume(TriclinicVolume):
    """
    An orthorhombic volume, a special case of a triclinic volume with ``alpha=beta=gamma=pi/2``
    """

    def __init__(self, *args):
        if len(args) == 3:
            a = args[0]
            b = args[1]
            c = args[2]
            alpha = pi / 2
            beta = pi / 2
            gamma = pi / 2
            TriclinicVolume.__init__(self, a, b, c, alpha, beta, gamma)
        else:
            # cell vectors
            v1 = np.array([float(f) for f in args[0:3]])
            v2 = np.array([float(f) for f in args[3:6]])
            v3 = np.array([float(f) for f in args[6:]])
            TriclinicVolume.__init__(self, v1, v2, v3)

    def __repr__(self):
        return "ORTHORHOMBIC a=%f b=%f c=%f" % (self.a, self.b, self.c)

    def __str__(self):
        return "ORT %f %f %f" % (self.a, self.b, self.c)


class TetragonalVolume(TriclinicVolume):
    """
    A tetragonal volume, a special case of a triclinic volume with ``a=b`` and ``alpha=beta=gamma=pi/2``
    """

    def __init__(self, *args):
        if len(args) == 2:
            a = args[0]
            c = args[1]
            b = a
            alpha = pi / 2
            beta = pi / 2
            gamma = pi / 2
            TriclinicVolume.__init__(self, a, b, c, alpha, beta, gamma)
        else:
            # cell vectors
            v1 = np.array([float(f) for f in args[0:3]])
            v2 = np.array([float(f) for f in args[3:6]])
            v3 = np.array([float(f) for f in args[6:]])
            TriclinicVolume.__init__(self, v1, v2, v3)

    def __repr__(self):
        return "TETRAGONAL a=%f c=%f" % (self.a, self.c)

    def __str__(self):
        return "TET %f %f" % (self.a, self.c)


class RhombohedralVolume(TriclinicVolume):
    """
    A rhombohedral volume, a special case of a triclinic volume with ``a=b=c`` and ``alpha=beta=gamma``.
    """

    def __init__(self, *args):
        if len(args) == 2:
            a = args[0]
            alpha = args[1]
            b = a
            c = a
            beta = alpha
            gamma = alpha
            TriclinicVolume.__init__(self, a, b, c, alpha, beta, gamma)
        else:
            # cell vectors
            v1 = np.array([float(f) for f in args[0:3]])
            v2 = np.array([float(f) for f in args[3:6]])
            v3 = np.array([float(f) for f in args[6:]])
            TriclinicVolume.__init__(self, v1, v2, v3)

    def __repr__(self):
        return "RHOMBOHEDRAL a=%f alpha=%f" % (self.a, self.alpha)

    def __str__(self):
        return "RHO %f %f" % (self.a, self.alpha)


class CubicVolume(TriclinicVolume):
    """
    A cubic volume, a special case of a triclinic volume with ``a=b=c`` and ``alpha=beta=gamma=pi/2``
    """

    def __init__(self, *args):
        if len(args) == 1:
            a = args[0]
            b = a
            c = a
            alpha = pi / 2
            beta = pi / 2
            gamma = pi / 2
            TriclinicVolume.__init__(self, a, b, c, alpha, beta, gamma)
        else:
            # cell vectors
            v1 = np.array([float(f) for f in args[0:3]])
            v2 = np.array([float(f) for f in args[3:6]])
            v3 = np.array([float(f) for f in args[6:]])
            TriclinicVolume.__init__(self, v1, v2, v3)

    def __repr__(self):
        return "CUBIC a=%f" % self.a

    def __str__(self):
        return "CUB %f" % self.a


class Volume(object):
    """
    Can create Subclasses from descriptive String
    """

    volumes = {
        "HEX": (HexagonalVolume, "ff"),
        "MON": (MonoclinicVolume, "ffff"),
        "TRI": (TriclinicVolume, "ffffff"),
        "ORT": (OrthorhombicVolume, "fff"),
        "TET": (TetragonalVolume, "ff"),
        "RHO": (RhombohedralVolume, "ff"),
        "CUB": (CubicVolume, "f"),
    }

    convert_functions = {"f": float, "i": int, "s": str}

    @classmethod
    def fromstring(cls, s):
        try:
            s = s.split()
            t = s[0].upper()  # volume type
            cl = cls.volumes[t][0]  # volume class
            if len(s) == 10:  # cell vectors given
                param = [float(f) for f in s[1:]]
            else:
                param_list = s[1:]
                # parsing parameter
                param = [cls.convert_functions[p](param_list[i]) for i, p in enumerate(cls.volumes[t][1])]
            return cl(*param)
        except Exception:
            return None



import os
import platform

import gr
import numpy as np
from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext as build_ext_orig


# Modified version of <https://stackoverflow.com/a/34830639/5958465>
class build_ext(build_ext_orig):
    def get_export_symbols(self, ext):
        # For CTypes extensions, do not require to export a `PyInit_` function
        if isinstance(ext, CTypes):
            return ext.export_symbols
        return super().get_export_symbols(ext)

    def get_ext_filename(self, fullname):
        # For CTypes extensions, force to use the default system prefix and extension for shared libraries.
        # This avoids file extensions like `.cpython-312-x86_64-linux-gnu.so`.
        if isinstance(self.ext_map[fullname], CTypes):
            shared_lib_prefix = {
                "Windows": "",
            }.get(platform.system(), "lib")
            shared_lib_ext = {
                "Darwin": ".dylib",
                "Windows": ".dll",
            }.get(platform.system(), ".so")
            fullname_components = fullname.split(".")
            ext_fullname = "".join(
                (
                    (
                        os.path.join(*fullname_components[:-1]),
                        os.path.sep,
                    )
                    if len(fullname_components) > 1
                    else tuple()
                )
                + (
                    shared_lib_prefix,
                    fullname_components[-1],
                    shared_lib_ext,
                )
            )
            return ext_fullname
        return super().get_ext_filename(fullname)


class CTypes(Extension):
    pass


gr_dir = os.path.dirname(os.path.dirname(gr.__gr._name))

setup(
    cmdclass={"build_ext": build_ext},
    ext_modules=[
        Extension(
            name="pymoldyn.computation.split_and_merge.util.numpy_extension.find_index_of_first_element_not_equivalent",
            sources=[
                "pymoldyn/computation/split_and_merge/util/numpy_extension/find_index_of_first_element_not_equivalent.c"
            ],
            include_dirs=[np.get_include()],
        ),
        Extension(
            name="pymoldyn.computation.split_and_merge.domain_centers.calculate_domain_centers",
            sources=["pymoldyn/computation/split_and_merge/domain_centers/calculate_domain_centers.c"],
            include_dirs=[np.get_include()],
        ),
        CTypes(
            name="pymoldyn.core.calculation.extension.algorithm",
            sources=["pymoldyn/core/calculation/extension/algorithm.c"],
            include_dirs=[os.path.join(gr_dir, "include")],
            library_dirs=[os.path.join(gr_dir, "lib")],
            libraries=["libGR3" if platform.system() == "Windows" else "GR3"],
        ),
    ],
)



# -*- coding: utf-8 -*-
"""
Calculate pair distribution functions.
"""


MINDISTANCE = 2.0  # shorter distances are ignored
PDFCUTOFF = 1.0  # g(r) = 0 if r < PDFCUTOFF


__all__ = ["PDF"]


import numpy as np
import math
from ..util.logger import Logger
import sys
from ..core.calculation.discretization import DiscretizationCache
import itertools
import os.path
from ..config.configuration import config


logger = Logger("statistics.pdf")
logger.setstream("default", sys.stdout, Logger.DEBUG)


class PDF(object):
    """
    Calculate pair distribution functions for atoms and cavities.
    """

    def __init__(self, *args):
        """
        Create a sample from atom and cavity positions and smooth them to
        get the PDFs

        The constructor can be called in two ways:

        - ``PDF(results)`` :
            retrieve the data from :class:`core.data.Results`

        - ``PDF(positions, elements, cavitycenters, volume)`` :
            use the given arrays and the volume object
        """
        if len(args) == 1:
            results = args[0]
            positions = results.atoms.positions
            elements = results.atoms.elements
            volume = results.atoms.volume
            if results.domains is not None:
                centers = results.domains.centers
                cachedir = os.path.expanduser(config.Path.cache_dir)
                cachepath = os.path.join(cachedir, "discretization_cache.hdf5")
                dcache = DiscretizationCache(cachepath)
                disc = dcache.get_discretization(volume, results.resolution)
                centers = list(map(disc.discrete_to_continuous, centers))
            else:
                centers = []
        elif len(args) == 4:
            positions, elements, centers, volume = args
        else:
            raise TypeError("PDF expects 1 or 4 parameters")

        self.positions = np.asarray(positions)
        self.elements = np.asarray(elements, dtype="|S4")
        self.centers = np.asarray(centers)
        self.volume = volume
        self.num_atoms = np.where(self.elements != "cav")[0].size
        self.numberdensity = float(self.num_atoms) / self.volume.volume

        self.stats = self._genstats(self.positions, self.elements, self.centers, self.volume)

    def pdf(self, elem1, elem2, cutoff=None, h=None, kernel=None):
        """
        Calculate a smoothed pair distribution function between the elements
        `elem1` and `elem2`.

        **Parameters:**
            `elem1`, `elem2` :
                Chemical element, e.g. 'Ge', 'Te' or 'cav' for cavities

            `h` :
                Smoothing parameter. The greater `h` is,
                the more the function is smoothed.

            `kernel` :
                Smoothing kernel

        **Returns:**
            Python function that represents the pair distribution function.
            It also accepts Numpy arrays as input.
            Returns `None` if the given elements do not exist or if there is
            not enough data to create the function.
        """
        if kernel is None:
            # kernel = Kernels.epanechnikov
            kernel = Kernels.epanechnikov

        data = None
        for s in self.stats:
            if set((elem1.lower(), elem2.lower())) == set((s[0].lower(), s[1].lower())):
                data = s[2]
                break
        if data is None:
            logger.debug("No statistical data for '{}-{}' found.".format(elem1, elem2))
            raise Exception("No statistical data for '{}' found.".format(elem1, elem2))

        if cutoff is None:
            cutoff = data.max()
        sel = np.where(np.logical_and(data > MINDISTANCE, data <= cutoff))[0]
        sel = data[sel]
        if h == 0:
            return sel
        if len(sel) < 2:
            logger.debug("Not enough data for '{}-{}' in cutoff={} range.".format(elem1, elem2, cutoff))
            raise Exception("Not enough data for '{}' in cutoff={} range.".format(elem1, elem2, cutoff))

        if h is None:
            ## magic constant
            ## minimizes the first peak
            # h = min(0.5, 0.5772778 * sel.min())
            h = 0.4

        # if h > 0.9 * sel.min():
        #    logger.debug("Bandwidth {} above threshold. Setting to {}.".format(h, 0.9 * sel.min()))
        #    h = 0.9 * sel.min()
        kde = Functions.KDE(sel, h=h, kernel=kernel)

        def wfunc(r):
            y = np.zeros_like(r)
            i = np.where(np.abs(r) > PDFCUTOFF)[0]
            y[i] = self.volume.volume / (data.size * 4 * math.pi * r[i] ** 2)
            return y

        return Functions.Product(kde, wfunc)

    @staticmethod
    def _correlatedistance(pos1, pos2=None, volume=None):
        """
        Measure the distances between coordinates.

        - ``_correlatedistance(pos1)`` :
            Correlate the coordinates in `pos1` with each other

        - ``_correlatedistance(pos1, pos2)`` :
            Correlate each coordinate in pos1 with each coordinate in pos2

        **Returns:**
            A sorted sample.
        """
        if volume is None:
            distance = lambda x, y: np.abs(y - x)
        else:
            distance = volume.get_distance

        if pos2 is None:
            n = pos1.shape[0]
            if n < 2:
                return np.zeros(0)
            a1 = np.vstack([pos1[0:-i, :] for i in range(1, n)])
            a2 = np.vstack([pos1[i:, :] for i in range(1, n)])
        else:
            n1 = pos1.shape[0]
            n2 = pos2.shape[0]
            a1 = np.hstack([pos1] * n2).reshape(n1 * n2, 3)
            a2 = np.vstack([pos2] * n1)
        samples = np.linalg.norm(distance(a1, a2), 2, axis=-1)
        samples.sort()

        return samples

    @classmethod
    def _genstats(cls, positions, elements, centers, volume=None):
        """
        Iterate through the elements and cavities and correlate the
        distances between atoms of that element/cavity and atoms of
        the other elements.
        """
        elemlist = np.unique(elements).tolist()
        if len(centers) > 0 and not "cav" in elemlist:
            elemlist.append("cav")
        pos = [None] * len(elemlist)
        for i, e in enumerate(elemlist):
            pos[i] = positions[np.where(elements == e)[0], :]

        if len(centers) > 0:
            i = elemlist.index("cav")
            if pos[i].size > 0:
                pos[i] = np.vstack((pos[i], centers))
            else:
                pos[i] = centers

        stats = []
        for i in range(len(elemlist)):
            try:
                elemlist[i] = elemlist[i].decode("utf-8")
            except AttributeError:
                pass
        for i, e1 in enumerate(elemlist):
            for j in range(i, len(elemlist)):
                e2 = elemlist[j]
                if i == j:
                    s = cls._correlatedistance(pos[i], volume=volume)
                else:
                    s = cls._correlatedistance(pos[i], pos[j], volume=volume)
                if len(s) > 1:
                    stats.append((e1, e2, s))

        return stats


class Functions(object):
    """
    Utility class with Callables
    """

    class KDE(object):
        """
        Kernel density estimation. Calculate a convolution
        of delta pulses with a smoothing kernel.
        """

        def __init__(self, x, y=None, h=1.0, kernel=None):
            if y is None:
                y = np.ones_like(x)
            if kernel is None:
                kernel = Kernels.gauss
            self.x = x
            self.y = y
            self.h = h
            self.kernel = kernel

        def __call__(self, p):
            result = np.zeros_like(p)
            if len(self.x) <= len(p):
                p = np.asarray(p)
                for xi, yi in zip(self.x, self.y):
                    result += yi * self.kernel((p - xi) / self.h) / self.h
            else:
                x = np.asarray(self.x)
                y = np.asarray(self.y)
                for i, pi in enumerate(p):
                    result[i] = np.sum(y * self.kernel((pi - x) / self.h) / self.h)

            return result

    class Product(object):
        """
        Product of two functions
        """

        def __init__(self, f, g):
            self.f = f
            self.g = g

        def __call__(self, x):
            return self.f(x) * self.g(x)


class Kernels(object):
    """
    Utility class with smoothing kernels
    """

    @staticmethod
    def gauss(x):
        c = 1.0 / math.sqrt(2.0 * math.pi)
        return c * np.exp(-0.5 * x**2)

    @staticmethod
    def compact(x):
        c = 2.25228362104
        if not isinstance(x, np.ndarray):
            if abs(x) < 1.0:
                return c * math.exp(1.0 / (x**2 - 1.0))
            else:
                return 0.0
        else:
            i = np.where(np.abs(x) < 1.0)[0]
            y = np.zeros(x.size)
            y[i] = c * np.exp(1.0 / (x[i] ** 2 - 1.0))
            return y

    @staticmethod
    def triang(x):
        c = 1.0
        if not isinstance(x, np.ndarray):
            if abs(x) < 1.0:
                return c * (1.0 - abs(x))
            else:
                return 0.0
        else:
            i = np.where(np.abs(x) < 1.0)[0]
            y = np.zeros(x.size)
            y[i] = c * (1.0 - np.abs(x[i]))
            return y

    @staticmethod
    def quad(x):
        c = 0.5
        if not isinstance(x, np.ndarray):
            if abs(x) < 1.0:
                return c
            else:
                return 0.0
        else:
            i = np.where(np.abs(x) < 1.0)[0]
            y = np.zeros(x.size)
            y[i] = c
            return y

    @staticmethod
    def posquad(x):
        c = 1.0
        if not isinstance(x, np.ndarray):
            if 0 <= x < 1.0:
                return c
            else:
                return 0.0
        else:
            i = np.where(np.logical_and(x >= 0.0, x < 1.0))[0]
            y = np.zeros(x.size)
            y[i] = c
            return y

    @staticmethod
    def negquad(x):
        c = 1.0
        if not isinstance(x, np.ndarray):
            if -1.0 < x <= 0.0:
                return c
            else:
                return 0.0
        else:
            i = np.where(np.logical_and(x > -1.0, x <= 0.0))[0]
            y = np.zeros(x.size)
            y[i] = c
            return y

    @staticmethod
    def epanechnikov(x):
        if not isinstance(x, np.ndarray):
            if abs(x) < 1.0:
                return 3 / 4 * (1.0 - x**2)
            else:
                return 0.0
        else:
            i = np.where(np.abs(x) < 1.0)[0]
            y = np.zeros_like(x)
            y[i] = 3.0 / 4.0 * (1.0 - x[i] ** 2)
            return y

    @staticmethod
    def bandwidth(n, d=1):
        """
        Scott's factor for bandwidth estimation
        """
        return (4.0 / 3.0 * n) ** (-1.0 / (d + 4.0))


class _TestPDF(object):
    @staticmethod
    def continuous_coordinates(coords, volume, resolution):
        cachedir = os.path.expanduser(config.Path.cache_dir)
        cachepath = os.path.join(cachedir, "discretization_cache.hdf5")
        dcache = DiscretizationCache(cachepath)
        disc = dcache.get_discretization(volume, resolution)

        return np.array(map(disc.discrete_to_continuous, coords))

    @staticmethod
    def plotfunc(pdf, e1, e2, px, h, *args):
        gr = pdf.pdf(e1, e2, h=h)
        plt.plot(px, gr(px), *args, label=str(h))
        py = gr(px)
        m = np.argmax(py)
        print("h={}, xi={}, g({}) = {}".format(gr.f.h, gr.f.x.min(), px[m], py[m]))

    @classmethod
    def plotpdf(cls, pdf, e1, e2):
        px = np.linspace(0, 2, 1000)
        plt.figure()
        cls.plotfunc(pdf, e1, e2, px, 0.25, "g--")
        cls.plotfunc(pdf, e1, e2, px, 0.5, "r-")
        # cls.plotfunc(pdf, e1, e2, px, 1.0, "b--")
        plt.legend(loc=0)
        plt.title("{}-{}".format(e1, e2))

    @classmethod
    def run(cls):
        import core.calculation as calculation

        calc = calculation.Calculation("../results")
        filename = "../xyz/structure_c.xyz"
        resolution = 64
        frame = 9
        # filename = "../xyz/GST_111_128_bulk.xyz"
        # resolution = 256
        # frame = 0
        settings = calculation.CalculationSettings({filename: [frame]}, resolution, True, False, False)
        print("calculating...")
        res = calc.calculate(settings)[0][0]
        print("generating statistics...")
        pdf = PDF(res)
        # centers = cls.continuous_coordinates(res.domains.centers,
        #                                     res.atoms.volume,
        #                                     res.resolution)
        # pdf = PDF(res.atoms.positions, res.atoms.elements,
        #                  centers, res.atoms.volume)

        print("plotting...")
        # cls.plotpdf(pdf, "Ge", "Ge")
        # cls.plotpdf(pdf, "Ge", "Te")
        cls.plotpdf(pdf, "cav", "cav")
        plt.show()


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    _TestPDF.run()
    # x = np.linspace(-2, 2, 200)
    # plt.plot(x, Kernels.gauss(x))
    # plt.plot(x, Kernels.compact(x))
    # plt.plot(x, Kernels.triang(x))
    # plt.plot(x, Kernels.quad(x))
    # plt.plot(x, Kernels.epanechnikov(x))
    # plt.show()



# -*- coding: utf-8 -*-

from math import floor
import collections
import itertools
import numpy as np
import h5py
from ...util.message import print_message, progress, finish
from . import algorithm
from .extension import mark_translation_vectors

try:
    import numexpr as ne

    NUMEXPR = True
except ImportError:
    NUMEXPR = False


__all__ = ["DiscretizationCache", "AtomDiscretization"]

dimension = algorithm.dimension
dimensions = algorithm.dimensions


class DiscretizationCache(object):
    """
    Instances of this class use a hdf5-formatted file as a cache for volume
    discretizations.
    """

    def __init__(self, filename):
        self.filename = filename
        self.file = None

    def lock(self):
        try:
            self.file = h5py.File(self.filename, "a")
            if "/discretizations" not in self.file:
                self.file.create_group("/discretizations")
        except IOError:
            self.file = None
        return self

    def unlock(self):
        if self.file is not None:
            self.file.close()
            self.file = None

    def __enter__(self):
        return self.lock()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.unlock()

    def get_discretization(self, volume, d_max):
        """
        Return a cached discretization are generate a new one, store it in the
        cache file and then return it.
        """
        discretization_repr = repr(volume) + " d_max=%d" % d_max
        print_message(
            "{volume}, discretization resolution: {resolution:d}".format(volume=repr(volume), resolution=d_max)
        )
        if self.file is not None and discretization_repr in self.file["/discretizations"]:
            stored_discretization = self.file["/discretizations/" + discretization_repr]
            grid = np.array(stored_discretization)
            discretization = Discretization(volume, d_max, grid)
        else:
            discretization = Discretization(volume, d_max)
            grid = discretization.grid
            if self.file is not None:
                self.file["/discretizations/" + discretization_repr] = grid
                self.file.flush()
        return discretization

    def get_discretization_from_string(self, string):
        string_parts = string.split()
        volume_type = string_parts[0]
        volume_type = volume_type.title() + "Volume"
        volume_class = getattr(volumes, volume_type)
        argument_info = string_parts[1:-1]
        arguments = {}
        for argument in argument_info:
            name, value = argument.split("=")
            value = float(value)
            arguments[name] = value
        volume = volume_class(**arguments)
        d_max_info = string_parts[-1]
        if not d_max_info.startswith("d_max="):
            raise Exception('Any volume discretization information string must end with "d_max=<d_max>".')
        d_max = float(d_max_info[len("d_max=") :])
        return self.get_discretization(volume, d_max)


class Discretization(object):
    """
    Instances of this class represent a discrete version of a volume.

    The discretization is done in the following steps:
     1. For a given maximum resolution of the resulting discrete representation
        (d_max), the length of the discretization cells (s_step) is calculated.
     2. The actual resolutions of the resulting discrete representation (d) are
        calculated and the side lengths of corresponding bounding cuboid
        (s_tilde) are calculated. (After this point, the transformation
        functions discrete_to_continuous(p) and continuous_to_discrete(p) can
        be used.)
     3. The volume's translation vectors are discretized (translation_vectors)
        and combinations of these are calculated
        (combined_translation_vectors).
     4. An integer grid is created and for each discrete point, it stores
        either 0 if the point is inside the volume or 1 if it is outside the
        volume.
     5. In the grid, for each point which is inside the volume (value = 0) all
        points created by addition of a combined translation vector are defined
        to be outside of the volume (value = 1), except for the point itself.
        After this, there is no equivalent pair of points, which both are
        inside the volume.
     6. At this point there might still be some points in the grid though,
        which have no equivalent point inside of the volume (value = 0). For
        each of these points, all equivalent points are found and from these,
        the one closest to the center (continuous coordinates (0,0,0)) is
        defined to be inside the volume.
     7. For each point which is outside of the volume (value = 1). A negative
        value is set to indicate which combined translation vector leads to the
        equivalent point inside the volume.
    """

    def __init__(self, volume, d_max, grid=None):
        # step 1
        self.d_max = d_max
        self.volume = volume
        self.s = volume.side_lengths
        self.s_max = max(self.s)
        if self.d_max % 2 == 1:
            self.d_max -= 1
        self.s_step = self.s_max / (self.d_max - 4)
        print_message("Step length:", self.s_step)

        # step 2
        dimensions = range(len(self.s))
        self.d = [int(floor(self.s[i] / self.s_step) + 4) for i in dimensions]
        for i in dimensions:
            if self.d[i] % 2 == 1:
                self.d[i] += 1
        print_message(
            "Resolution per axis: (x: {x:d}, y: {y:d}, z: {z:d})".format(x=self.d[0], y=self.d[1], z=self.d[2])
        )
        self.s_tilde = [(self.d[i] - 1) * self.s_step for i in dimensions]

        # step 3
        self.translation_vectors = [
            [int(floor(c / self.s_step + 0.5)) for c in v] for v in self.volume.translation_vectors
        ]
        # TODO: remove unnecessary vectors in hexagonal volumes
        self.combined_translation_vectors = [
            [sum([v[0][j] * v[1] for v in zip(self.translation_vectors, i)]) for j in dimensions]
            for i in itertools.product((-1, 0, 1), repeat=len(dimensions))
            if any(i)
        ]

        if grid is not None:
            self.grid = grid
        else:
            # step 4
            self.grid = np.zeros(self.d, dtype=np.int8)
            # create array similar to itertools.product
            order = [i + 1 for i in range(len(dimensions))] + [0]
            points = np.indices(self.d).transpose(*order).reshape((-1, len(dimensions)))
            # choose points that are not inside
            inside = self.volume.is_inside(self.discrete_to_continuous(points))
            outside_points = points.compress(np.logical_not(inside), axis=0)
            del points
            # convert to a tuple of indices
            indices = tuple([outside_points[:, i] for i in range(len(dimensions))])
            self.grid[indices] = 1
            del outside_points
            del inside
            del indices
            # steps 5, 6, 7
            mark_translation_vectors(self.grid, self.combined_translation_vectors)
        translation_vector_output = ", ".join(["({0}, {1}, {2})".format(*vec) for vec in self.translation_vectors])
        print_message("Translation vectors:", translation_vector_output)

    def get_direct_neighbors(self, point):
        """
        This method returns the direct neighbor points of a given point.
        """
        neighbors_without_bound_checks = [
            [point[j] + i[j] for j in dimensions] for i in itertools.product((-1, 0, 1), repeat=dimension) if any(i)
        ]
        direct_neighbors = [
            tuple(neighbor)
            for neighbor in neighbors_without_bound_checks
            if all([0 <= neighbor[j] <= self.d[j] - 1 for j in dimensions])
        ]
        return direct_neighbors

    def get_neighbors_in_volume(self, point):
        """
        This method returns the neighbor points of a given point, which are
        inside the volume.
        """
        direct_neighbors = self.get_direct_neighbors(point)
        neighbors_in_volume = map(self.get_equivalent_point_in_volume, direct_neighbors)
        return neighbors_in_volume

    def get_equivalent_point_in_volume(self, point):
        """
        For a point given in discrete coordinates, this method returns the
        point which is inside of the volume and is equivalent to the given
        point.
        """
        # If the point cannot be translated directly, shift it in the direction of the cell center by
        # applying one of the discrete translation vectors in each step until the point is inside the
        # discretization grid
        cell_center = None
        while not all(0 <= p < s for p, s in zip(point, self.grid.shape)):
            if cell_center is None:
                cell_center = np.array(self.grid.shape, dtype=np.int32) // 2
                translation_vectors = np.array(self.translation_vectors)
                normalized_translation_vectors = (translation_vectors.T / np.linalg.norm(translation_vectors, axis=1)).T
            vector_to_cell_center = cell_center - np.array(point, dtype=np.int32)
            vector_to_cell_center = vector_to_cell_center / np.linalg.norm(vector_to_cell_center)
            dot_products = np.dot(normalized_translation_vectors, vector_to_cell_center)
            rounded_dot_products = np.array(np.round(dot_products), dtype=np.int32)
            needed_translation_vector_index = np.argmax(np.abs(dot_products))
            needed_translation_vector = (
                rounded_dot_products[needed_translation_vector_index]
                * translation_vectors[needed_translation_vector_index]
            )
            next_point = tuple(p + t for p, t in zip(point, needed_translation_vector))
            if next_point == point:
                raise ArithmeticError("Could not determine an equivalent point inside the volume")
            point = next_point
        if self.grid[point] == 0:
            return point
        else:
            combined_translation_vector_index = -self.grid[point] - 1
            combined_translation_vector = self.combined_translation_vectors[combined_translation_vector_index]
            return tuple([point[i] + combined_translation_vector[i] for i in dimensions])

    def get_translation_vector(self, point):
        combined_translation_vector_index = -self.grid[point] - 1
        combined_translation_vector = self.combined_translation_vectors[combined_translation_vector_index]
        return combined_translation_vector

    def continuous_to_discrete(self, arg, result_inside_volume=False, unit_exponent=1):
        """
        Transforms a single value or a point from continuous to discrete coordinates.
        If a point is given, ``result_inside_volume`` can be set to shift the result into the volume cell if
        necessary (equivalent point).
        In case of a single value input, 'unit_exponent' is used as exponent of the scaling factor.
        """

        def transform_point(point, result_inside_volume):
            if isinstance(point, np.ndarray) and len(point.shape) > 1:
                s_tilde_bc = np.tile(self.s_tilde, (point.shape[0], 1))
                print("s_tilde_bc", s_tilde_bc)
                result = np.array(
                    np.floor((point + s_tilde_bc / 2) / self.s_step + 0.5),
                    dtype=np.int32,
                )
                if result_inside_volume:
                    result = np.array(self.get_equivalent_point_in_volume(result))
            else:
                result = tuple(int(floor((point[i] + self.s_tilde[i] / 2) / self.s_step + 0.5)) for i in dimensions)
                if result_inside_volume:
                    result = self.get_equivalent_point_in_volume(result)
            return result

        def transform_value(value, unit_exponent):
            return int(floor(value / (self.s_step**unit_exponent) + 0.5))

        if isinstance(arg, collections.abc.Iterable):
            return transform_point(arg, result_inside_volume)
        else:
            return transform_value(arg, unit_exponent)

    def discrete_to_continuous(self, arg, result_inside_volume=False, unit_exponent=1):
        """
        Transforms a single value or a point from discrete to continuous coordinates.
        If a point is given, ``result_inside_volume`` can be set to shift the result into the volume cell if
        necessary (equivalent point).
        In case of a single value input, 'unit_exponent' is used as exponent of the scaling factor.
        """

        def transform_point(point):
            if isinstance(point, np.ndarray) and len(point.shape) > 1:
                if result_inside_volume:
                    if any(isinstance(c, float) for c in point):
                        rounded_point = np.array(np.around(point), dtype=np.int32)
                        rounded_diff = point - rounded_point
                        rounded_point = np.array(self.get_equivalent_point_in_volume(rounded_point))
                        point = rounded_point + rounded_diff
                    else:
                        point = np.array(self.get_equivalent_point_in_volume(point))
                if NUMEXPR:
                    s_tilde_bc = np.tile(self.s_tilde, (point.shape[0], 1))
                    s_step = self.s_step
                    return ne.evaluate("point * s_step - s_tilde_bc / 2")
                else:
                    s_tilde_bc = np.tile(self.s_tilde, (point.shape[0], 1))
                    return point * self.s_step - s_tilde_bc / 2
            else:
                if result_inside_volume:
                    if any(isinstance(c, float) for c in point):
                        rounded_point = tuple(int(round(c)) for c in point)
                        rounded_diff = tuple(c1 - c2 for c1, c2 in zip(point, rounded_point))
                        rounded_point = self.get_equivalent_point_in_volume(rounded_point)
                        point = tuple(c1 + c2 for c1, c2 in zip(rounded_point, rounded_diff))
                    else:
                        point = self.get_equivalent_point_in_volume(point)
                return tuple(point[k] * self.s_step - self.s_tilde[k] / 2 for k in dimensions)

        def transform_value(value, unit_exponent):
            return value * (self.s_step**unit_exponent)

        if isinstance(arg, collections.abc.Iterable):
            return transform_point(arg)
        else:
            return transform_value(arg, unit_exponent)

    def __repr__(self):
        return repr(self.volume) + " d_max=%d" % self.d_max


class AtomDiscretization:
    """
    Instances of this class represent a list of atoms in a discrete volume.
    """

    def __init__(self, atoms, discretization):
        self.atoms = atoms
        self.discretization = discretization
        self.sorted_discrete_radii = []
        self.discrete_radii = []
        self.discrete_positions = []
        for radius_index, position in zip(self.atoms.radii_as_indices, self.atoms.sorted_positions):
            radius = self.atoms.sorted_radii[radius_index]
            discrete_position = self.discretization.continuous_to_discrete(position)
            self.discrete_positions.append(discrete_position)
        for radius in self.atoms.sorted_radii:
            discrete_radius = int(floor(radius / self.discretization.s_step + 0.5))
            self.sorted_discrete_radii.append(discrete_radius)
        print_message("Maximum radius:", self.sorted_discrete_radii[0])



import faulthandler
import importlib
import io
import os
import signal
import sys


def start_batch():
    # The user actually wants to use startBatch.py instead of startGUI.py
    # This is implemented here as both Linux and OS X use completely
    # different ways to implement the pymoldyn executable, but both call
    # startGUI.py.
    from ..cli import main as cli_main

    if sys.argv[1] == "--batch":
        del sys.argv[1]
    cli_main()


def test_gui_dependencies():
    gui_dependencies = {
        "gr3": "gr",
        "OpenGL": "PyOpenGL",
        "PySide6": "PySide6",
    }

    missing_dependencies = []
    for dep in gui_dependencies:
        try:
            importlib.import_module(dep)
        except ImportError:
            missing_dependencies.append(dep)
    if missing_dependencies:
        print(
            "The following GUI dependencies are missing: "
            + f"{', '.join(gui_dependencies[dep] for dep in missing_dependencies)}",
            file=sys.stderr,
        )
        print(
            "Please install them (or `pip install pymoldyn[gui]`) and try again.",
            file=sys.stderr,
        )
        return False
    return True


def start_gui():
    from PySide6 import QtCore, QtGui, QtWidgets

    from ..core import file
    from ..core.control import Control
    from . import main_window

    # By default, GR3 uses software rendering. Set `GR3_USE_OPENGL=1` to activate hardware accelerated OpenGL rendering.
    # This is needed, otherwise the Qt OpenGL widget cannot work.
    # TODO: Try software rendering without an OpenGL widget since the software renderer often has even better
    # performance.
    os.environ["GR3_USE_OPENGL"] = "1"

    file.SEARCH_PATH = os.getcwd()
    os.chdir(os.path.dirname(os.path.realpath(__file__)))

    app = QtWidgets.QApplication(sys.argv)
    control = Control()
    window = main_window.MainWindow(control)

    app.setOrganizationName("Forschungszentrum Jülich GmbH")
    app.setOrganizationDomain("fz-juelich.de")
    app.setApplicationName("pyMolDyn")

    # Let the Python interpreter run every 50ms...
    timer = QtCore.QTimer()
    timer.start(50)
    timer.timeout.connect(lambda: None)
    # ... to allow it to quit the application on SIGINT (Ctrl-C)
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    filenames = sys.argv[1:]
    for filename in filenames:
        if filename:
            window.file_dock.file_tab.disable_files_in_menu_and_open(filename)
            window.update_submenu_recent_files()
    screen_geom = window.screen().availableGeometry()
    window.setGeometry(
        (screen_geom.width() - window.width()) / 2,
        (screen_geom.height() - window.height()) / 2,
        window.width(),
        window.height(),
    )
    sys.exit(app.exec())


def main():
    try:
        # NSLog on macOS raises an `io.UnsupportedOperation` when `fileno` is called
        # `faulthandler` raises RuntimeError if `sys.stderr` is `None` (`noconsole` mode on Windows)
        faulthandler.enable(file=sys.stderr, all_threads=True)
    except (io.UnsupportedOperation, RuntimeError):
        pass
    if len(sys.argv) > 1 and sys.argv[1] == "--batch":
        start_batch()
    else:
        if not test_gui_dependencies():
            sys.exit(1)
        start_gui()


if __name__ == "__main__":
    main()
